from select import select
import cv2 as cv
from cv2 import add
import numpy as np
import math
from matplotlib import pyplot as plt 
import copy
from numpy.lib.function_base import append
from scipy import optimize
import time
from sklearn import preprocessing
# from torch import rand
import random
import functools
from numpy import *



def img_init(src, double = False):
    '''
    @ Function: 载入图像+预处理
    @ Input:  src->(Mat/Array)   彩色图像  
              double->bool=False 是否使用插值
    @ Output: Points->(Array)    轮廓点集合 
              pointnum->(int)    轮廓点数量  
              src->(Mat/Array)   轮廓图像
    @ Author: lxc
    @ Last modified time: 2022/08/21
    '''
    imgray = cv.cvtColor(src, cv.COLOR_BGR2GRAY)
    if double:
        imgray = cv.resize(imgray,dsize=(2*imgray.shape[1],2*imgray.shape[0]),interpolation=cv.INTER_LANCZOS4)
    # cv.imshow("1",imgray)
    # cv.waitKey(0)
    ret, thresh = cv.threshold(imgray, 100, 255, 0)
    cv.bitwise_not(thresh,thresh)
    # cv.imshow("1",thresh)
    # cv.waitKey(0)
#提取轮廓云点，创建画板
    contours, hierarchy = cv.findContours(thresh, cv.RETR_TREE,cv.CHAIN_APPROX_NONE)
    cnt = contours
    src = np.zeros((imgray.shape[0],imgray.shape[1],3), np.uint8)
    src.fill(255)
    src = cv.drawContours(src, contours, -1, (0,0,0),1)
    Points = np.array(cnt)
    pointnum = len(Points)
    return Points,pointnum,src




def get_anglesGroup(Ps, num, partmode = [-1,-1]):
    '''
    @ Function: 获得angle_,angle
    @ Input:  Ps->(Array)            轮廓点集合
              num->(int)             步长  
              partmode->list=[-1,-1] 默认截取全部轮廓点
    @ Output: AnglesGroup->(Array)   前弦倾角集合  
              Angles_Group->(Array)  后弦倾角集合
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    pointnum = len(Ps)
    AnglesGroup = np.empty(0,np.float16)
    Angles_Group = np.empty(0,np.float16)
    def get_angles(P, pointnum, i ,num):
        end = i-num
        angle_ = (P[end] - P[i])
        angle_ = np.arctan2(np.cross(angle_,np.array([1,0])), 
                            np.dot(angle_,np.array([1,0])))*180/np.pi
        if angle_<0:
            angle_ = angle_+360
        start = i+num
        if start>pointnum-1:
            start = start-pointnum
        angle = (P[i] - P[start])
        angle = np.arctan2(np.cross(angle,  np.array([1,0])), 
                        np.dot(angle,  np.array([1,0])))*180/np.pi
        if angle<0:
            angle = angle+360
        return angle_,angle
    #获得angle_,angle集合Angle、Angle_
    if partmode[0] == -1 and partmode[1] == -1 :
        part = [0, pointnum]
    else:
        part = partmode
    for i in range(int(part[0]),int(part[1]),1):
        angle_, angle = get_angles(Ps, pointnum, i, num)         
        AnglesGroup = np.append(AnglesGroup,angle)
        Angles_Group = np.append(Angles_Group,angle_)
    return AnglesGroup, Angles_Group



#
def get_roi(AnglesGroup, Angles_Group, newPoints = False , init_group = 0):
    '''
    @ Function: Angle、Angle_前后互补,获得特征点
    @ Input: AnglesGroup->(Array)  前弦倾角集合  
             Angles_Group->(Array) 后弦倾角集合
             newPoints->bool=Fales 默认不返回互补后的倾角集合
             init_group->int=0     默认点簇分组： 0
    @ Output: roi_point->(Array)   初筛特征点
              Point2->(Array)      [0]/互补后的倾角集合
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    pointnum = len(AnglesGroup)
    j,j1,j2 =0,0,0
    roi_point = np.empty(0,np.float16)
    Point2 =  np.empty(0,np.float16)
    while True:
        j = max(j1,j2)
        if j>=pointnum-1:
            break
        for j1 in range(j,pointnum,1):
            if AnglesGroup[j]!=AnglesGroup[j1] and AnglesGroup[j]!=Angles_Group[j1]:
                break
        for j2 in range(j,pointnum,1):
            if Angles_Group[j]!=AnglesGroup[j2] and Angles_Group[j]!=Angles_Group[j2]:
                break
        if j1>j2:
            num = AnglesGroup[j]
        else:
            num = Angles_Group[j]

        if j==0 or max(j1,j2)!=roi_point[-3]:
            Point2 = np.append(Point2,[num]*(max(j1,j2)-j))
            roi_point = np.append(roi_point,j)
            roi_point = np.append(roi_point,max(j1,j2))
            roi_point = np.append(roi_point,AnglesGroup[max(j1,j2)])
            roi_point = np.append(roi_point,init_group)
    roi_point = roi_point.reshape([-1,4])
    if newPoints:
        return roi_point, Point2
    else:
        return roi_point, np.empty(0,np.float16)




# # 点簇
def get_pointsgroups(pointnum, roi_point):
    '''
    @ Function: Angle、Angle_前后互补,获得特征点
    @ Input:  pointnum->(int)      轮廓点个数
              roi_point->(Array)   初筛特征点
    @ Output: roi_point->(Array)   分点簇后的初筛特征点
              roi_point_1->(list)  分点簇后的初筛特征点      
    @ Author: lxc 
    @ Last modified time: 2022/07/16
    '''
    roi_point_1 = [[]]
    sign = 1
    for roi in range(0,len(roi_point),1):
        if abs(roi_point[roi][1]-roi_point[roi-1][1])<5 or \
       abs(roi_point[roi][1]-roi_point[roi-1][1])>pointnum-6 :
            if roi_point[roi][3] ==0 and roi_point[roi-1][3] !=0:
                roi_point[roi][3]=roi_point[roi-1][3]
                roi_point_1[sign-1].append(roi_point[roi][1])
            elif roi_point[roi][3] ==0 and roi_point[roi-1][3] ==0:
                roi_point[roi][3]=sign
                roi_point[roi-1][3]=sign
                roi_point_1[sign-1].append(roi_point[roi][1])
                roi_point_1[sign-1].append(roi_point[roi-1][1])
            elif roi_point[roi][3] !=0 and roi_point[roi-1][3] !=0:
                roi_point_1[0].append(roi_point[roi-1][1])
                aa = roi_point[roi-1][3]
                for i in range(2,len(roi_point)+1,1):
                    if roi_point[-i][3] == aa:
                        roi_point[-i][3] = roi_point[roi][3]
                        roi_point_1[sign-1].remove(roi_point[-i][1])
                        roi_point_1[0].append(roi_point[-i][1])
                    else:
                        break
        else:
            if roi_point[roi][3] ==0:
                roi_point_1.append([])
                sign = sign+1
                roi_point[roi][3] = sign
                roi_point_1[sign-1].append(roi_point[roi][1])
            else:
                roi_point_1[0].append(roi_point[roi][1])
    if [] in roi_point_1:
        roi_point_1.remove([])
    # roi_point_1 = np.array(roi_point_1)
    return roi_point,roi_point_1

def adaptive_stepsize(Points,roi_point,roi_point_1):
    '''
    @ Function: 自适应步长
    @ Input:  Points->(Array)      轮廓点集合
              pointnum->(int)      轮廓点个数
              roi_point->(Array)   分点簇后的初筛特征点
              roi_point_1->(list)  分点簇后的初筛特征点 
    @ Output: roi_point->(Array)   自适应步长矫正后的特征点
              roi_point_1->(list)  自适应步长矫正后的特征点      
    @ Author: lxc 
    @ Last modified time: 2022/08/01
    '''
    pointnum = len(Points)
    def paixu(data):
        def totuple(a):
            try:
                return tuple(totuple(i) for i in a)
            except TypeError:
                return a
        m2A = []
        for i in range(0, len(data)):
            m2A.append((data[i][0], data[i][1], data[i][2], data[i][3]))
        dtype = [('x', float), ('y', float), ('z', float), ('i', int)]
        tuple1 = np.array(m2A, dtype)
        tuple1 = np.sort(tuple1, order=['y', 'i'])
        inFile = np.array(totuple(tuple1))   # 转为元组，转array
        return inFile

    p_num = 0
    init_group_ = roi_point[0][3]
    for i, ps in enumerate(roi_point_1):
        r_Pnum = len(roi_point)
        init_group = init_group_+i
        if len(ps) == 1 and p_num<r_Pnum-1 and p_num>0:
            p1 = roi_point[p_num-1][1]
            p1_len = roi_point[p_num][1]-roi_point[p_num-1][1]
            p2 = roi_point[p_num+1][1]
            p2_len = roi_point[p_num+1][1]-roi_point[p_num][1]
            # points_part = Points[p1:p2]
            stepsize_ = min(int(max(p1_len,p2_len)/50),min(p1_len,p2_len))
            stepsize = max(10,stepsize_)
            AG, A_G = get_anglesGroup(Points, num = stepsize, partmode = [p1,p2])
            roi_p, Point2 = get_roi( AG, A_G, init_group = init_group)
            for r_p in range(0,len(roi_p)):
                roi_p[r_p][1] = roi_p[r_p][1]+p1
                roi_p[r_p][0] = roi_p[r_p][0]+p1
                roi_point_1[i].append(roi_p[r_p][1])
                roi_point_1[i].sort()

            roi_point = np.append(roi_point,roi_p,axis=0)
            roi_point = paixu(roi_point)
            # roi_point = roi_point[np.argsort(roi_point[:,3])]
        p_num = p_num+len(roi_point_1[i])
    for r_p in roi_point:
       r_p[3] = 0    
    roi_point,roi_point_1 = get_pointsgroups(pointnum, roi_point)     
    return roi_point,roi_point_1



#
def Fitting_Line(x,y):
    '''
    @ Function: 直线最小二乘法拟合
    @ Input:  x->(Array)      轮廓点x集合
              y->(Array)      轮廓点y集合
    @ Output: Emax_line       直线最大误差    
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    method_1  = "Least Squares"
    point =  np.c_[x, y]
    output = cv.fitLine(point, cv.DIST_L2, 0, 0.01, 0.01)
    b = np.arctan(output[1] / output[0])
    r = np.cos(b)*output[3] - np.sin(b)* output[2]
    r_ = abs(np.cos(b)*y - np.sin(b)* x - r)
    Emax_line = np.max(r_)
    return Emax_line

 #圆弧最小二乘法拟合
def Fitting_Circle(x,y):
    '''
    @ Function: 圆弧最小二乘法拟合
    @ Input:  x->(Array)      轮廓点x集合
              y->(Array)      轮廓点y集合
    @ Output: Emax_line       圆弧最大误差    
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    method_2  = "Fitting Circle"
    # 质心坐标
    x_m = np.mean(x)
    y_m = np.mean(y)
    def calc_R(xc, yc):
        return np.sqrt((x - xc) ** 2 + (y - yc) ** 2)
    def f_2(c):
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    #圆心估计
    center_estimate = x_m, y_m
    center_2, _ = optimize.leastsq(f_2, center_estimate)
    xc_2, yc_2 = center_2
    Ri_2       = calc_R(xc_2, yc_2)
    #拟合圆的半径
    R_2        = Ri_2.mean()
    Emax_residu  = max((Ri_2 - R_2))
    if len(x)/R_2<np.pi/6:
        Emax_residu =max(1.2*(1-((len(x)/R_2)/np.pi))+Emax_residu,Emax_residu)
    # elif Emax_residu < 9:
    #     Emax_residu = 0.5
    return Emax_residu


def Fast_line_fit(x,y):
    '''
    @ Function: 自创直线快速拟合（没卵用）
    @ Input:  x->(Array)      轮廓点x集合
              y->(Array)      轮廓点y集合
    @ Output: Emax_line       直线最大误差    
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    ind = np.linspace(0,len(x)-1,5).astype(np.int16)
    #  （y-y1）/(y2-y1)-(x-x1)(x2-x1)=0
    X1 = x[ind[0]]
    X2 = x[ind[-1]]
    Y1 = y[ind[0]]
    Y2 = y[ind[-1]]
    # x = x[ind]
    # y = y[ind]
    if X1 == X2 :
        return np.max(abs(x-X1))
    elif Y1 == Y2:
        return np.max(abs(y-Y1))
    else:
        return np.max(abs((y-Y1)/(Y2-Y1)-(x-X1)/(X2-X1)))

def Fast_circle_fit_1(x,y):
    '''
    @ Function: 这个是矩方法拟合（误差较大）
    @ Input:  x->(Array)      轮廓点x集合
              y->(Array)      轮廓点y集合
    @ Output: Emax_line       圆弧最大误差    
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    x_m = mean(x)
    y_m = mean(y)
    def calc_R(xc, yc):
        return sqrt((x - xc) ** 2 + (y - yc) ** 2)
    def f_2(c):
        Ri = calc_R(*c)
        return Ri - Ri.mean()
    center_estimate = x_m, y_m
    center_2, _ = optimize.leastsq(f_2, center_estimate)
    xc_2, yc_2 = center_2
    Ri_2       = calc_R(xc_2, yc_2)
    R_2        = Ri_2.mean()
    residu_r  = max(Ri_2 - R_2)
    return residu_r



def Fast_circle_fit(x,y):
    '''
    @ Function: 自创圆弧快速拟合（没卵用）
    @ Input:  x->(Array)      轮廓点x集合
              y->(Array)      轮廓点y集合
    @ Output: Emax_line       圆弧最大误差    
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    ind = np.linspace(0,len(x)-1,5).astype(np.int16)
    #  （y-y1）/(y2-y1)-(x-x1)(x2-x1)=0
    X1 = x[ind[0]]
    X2 = x[ind[2]]
    X3 = x[ind[-1]]
    Y1 = y[ind[0]]
    Y2 = y[ind[2]]
    Y3 = y[ind[-1]]
    d1 = (X2**2+Y2**2)-(X1**2+Y1**2)
    d2 = (X3**2+Y3**2)-(X2**2+Y2**2)
    fm = 2*((Y3-Y2)*(X2-X1)-(Y2-Y1)*(X3-X2))
    Xo = ((Y3-Y2)*d1-(Y2-Y1)*d2)/fm
    Yo = ((X2-X1)*d2-(X3-X2)*d1)/fm
    Dm = (X2-Xo)**2+(Y2-Yo)**2
    # x = x[ind]
    # y = y[ind]
    return np.max(abs((x-Xo)**2+(y-Yo)**2 - Dm))


def get_x_y(Points, pointnum, part, n):
    '''
    @ Function: 获取一段轮廓点的x,y
    @ Input:  Points->(Array)      轮廓点集合
              pointnum->(int)      轮廓点个数
              part->(list)         轮廓段启、终点
              n->(int)             缩进长度
    @ Output: x,y                  横纵坐标列表  
    @ Author: lxc 
    @ Last modified time: 2022/07/15
    '''
    x = np.zeros(0)
    y = np.zeros(0)
    ind1 = int(part[0]+n)
    ind2 = int(part[1]-n)
    if ind1>ind2:
        for i in Points[ind1:pointnum-1]:
            x = np.append(x,i[0])
            y = np.append(y,i[1])
        for i in Points[0:ind2]:
            x = np.append(x,i[0])
            y = np.append(y,i[1])
    else:
        for i in Points[ind1:ind2]:
            x = np.append(x,i[0])
            y = np.append(y,i[1])
    # if len(x)>200:
    #     sel  = np.linspace(0, len(x)-1, 150, dtype=int)
    #     x = x[sel]
    #     y = y[sel]
    return x,y           

#获取待选特征点区域
def get_parts_grade(Points,pointnum, roi_point, T):
    '''
    @ Function: 轮廓段评分
    @ Input:  Points->(Array)      轮廓点集合
              pointnum->(int)      轮廓点个数
              roi_point->(list)    初筛轮廓点
              T->(int)             容错阈值     
    @ Output: roi_point_2          轮廓段信息表   
    @ Author: lxc 
    @ remarks: 此处直接使用计算好n值,具体实现见n_Otsu.py
    @ Last modified time: 2022/07/15
    '''
    Error = 0
    roi_point_2 = np.empty(0,np.float16)
    for roi in range(0,len(roi_point),1):
        # if lxc_num[2] != roi_point[roi][3] :
        #     lxc_num = [-1, -1, -1]
        #     lxc_num[2] = roi_point[roi][3]
        for  i in range(1,len(roi_point),1):
            if roi_point[roi-i][3]==roi_point[roi][3]:
                continue  #同一点簇的点不进行运算
            part = [roi_point[roi-i][1],roi_point[roi][1]]
            x,y=get_x_y(Points, pointnum, part,0)
            # if lxc_num[0] == roi_point[roi-i][3] :
            #     if lxc_num[1] ==1 :
            #         Error = Fitting_Circle(x,y)
            #     if lxc_num[1] ==2 :
            #         Error = Fitting_Line(x,y)
            # else:
            Emax_residu = Fitting_Circle(x,y)    
            Emax_line = Fitting_Line(x,y)
            Error = min(Emax_residu,Emax_line)
                # if Error == Emax_residu:
                #     lxc_num[1] =1
                # if Error == Emax_line:
                #     lxc_num[1] =2
                # lxc_num[0] = roi_point[roi-i][3]

            if Error>T:
                break
            roi_point_2 = np.append(roi_point_2,roi_point[roi][1])
            roi_point_2 = np.append(roi_point_2,roi_point[roi-i][1])
            roi_point_2 = np.append(roi_point_2,Error)
            roi_point_2 = np.append(roi_point_2,len(x))
            roi_point_2 = np.append(roi_point_2,roi_point[roi][3])
            roi_point_2 = np.append(roi_point_2,roi_point[roi-i][3])  
    roi_point_2 = roi_point_2.reshape([-1,6])   
    min_max_scaler = preprocessing.MinMaxScaler()
    emax_minmax = min_max_scaler.fit_transform(roi_point_2[:,2].reshape(-1, 1))
    lenth_minmax = min_max_scaler.fit_transform(roi_point_2[:,3].reshape(-1, 1))
# print(lenth_minmax)
    result = 20*lenth_minmax-emax_minmax
    roi_point_2 = np.append(roi_point_2, result, axis=1)
    return roi_point_2



def get_preciseROI(pointnum, roi_point_2):
    '''
    @ Function: NMS进行轮廓段筛选
    @ Input:  pointnum->(int)      轮廓点个数
              roi_point_2->(list)  评价后的轮廓段信息 
    @ Output: roi_point_3          筛选后的轮廓段端点   
    @ Author: lxc 
    @ Last modified time: 2022/08/30
    '''
    roi_point_3 = np.empty(0,np.int16)
    while(len(roi_point_2)>0):
        Select = [True]*len(roi_point_2)
        a=np.where(roi_point_2[:,6]==np.max(roi_point_2[:,6]))
        for i in a:
            Select[i[0]] = False
            roi_point_3 = np.append(roi_point_3,roi_point_2[i[0]][0])
            roi_point_3 = np.append(roi_point_3,roi_point_2[i[0]][1])
            j= 0
            a0 = roi_point_2[i[0]][0]
            b0 = roi_point_2[i[0]][1]
            if a0<b0:
                b0 = a0-roi_point_2[i[0]][3]
            while(True):
                j = j+1
                ind = i[0]+j
                if ind>len(roi_point_2)-1:
                    ind = ind-len(roi_point_2)
                if ind>len(roi_point_2)-1:
                    break
                a1 = roi_point_2[ind][0]
                b1 = roi_point_2[ind][1]
                if a1<b1:
                    b1 = a1-roi_point_2[ind][3]
                elif b0<0 and a1-pointnum>b0:
                    a1 = a1-pointnum
                    b1 = b1-pointnum
                cover = min(a0,a1)-max(b0,b1)
                if cover/min(roi_point_2[i[0]][3],roi_point_2[ind][3])>0.5:
                    Select[ind] = False
                else:
                    if a1>b0 and a1<a0:
                        roi_point_2[ind][0] = roi_point_2[i[0]][1]
                        if roi_point_2[ind][5] == roi_point_2[i[0]][5] :
                            Select[ind] = False
                    if b1>b0 and b1<a0:
                        roi_point_2[ind][1] = roi_point_2[i[0]][0]
                        if roi_point_2[ind][4] == roi_point_2[i[0]][4] :
                            Select[ind] = False

            j= 0
            while(True):
                j = j-1
                ind = i[0]+j
                if abs(i[0]+j )>len(roi_point_2):
                    break
                a1 = roi_point_2[ind][0]
                b1 = roi_point_2[ind][1]
                if a1<b1:
                    b1 = a1-roi_point_2[ind][3]
                elif b0<0 and a1-pointnum>b0:
                    a1 = a1-pointnum
                    b1 = b1-pointnum
                cover = min(a0,a1)-max(b0,b1)
                if cover/min(roi_point_2[i[0]][3],roi_point_2[ind][3])>0.5:
                    Select[ind] = False
                else:
                    if a1>b0 and a1<a0:
                        roi_point_2[ind][0] = roi_point_2[i[0]][1]
                        if roi_point_2[ind][5] == roi_point_2[i[0]][5] :
                            Select[ind] = False
                    if b1>b0 and b1<a0:
                        roi_point_2[ind][1] = roi_point_2[i[0]][0]
                        if roi_point_2[ind][4] == roi_point_2[i[0]][4] :
                            Select[ind] = False
        roi_point_2 = roi_point_2[Select]
    roi_point_3 = roi_point_3.reshape(-1,2)
    return roi_point_3

def write_data(P, name, Points, pointytpe=False):
    '''
    @ Function: 保存数据
    @ Input:  P->(list)            待保存数据
              rname->(string)      保存后文件名 
              Points->(list)       所有轮廓点坐标
              pointytpe—>(bool)    是否保存为点坐标
    @ Output: None                 无返回  
    @ Author: lxc 
    @ Last modified time: 2022/07/30
    '''
    fo = open("C:../data/"+name+".txt", "w")
    if len(P)>1:
        for i in P:
            if pointytpe:
                point = Points[int(i)]
                y=str(point)
            else:
                y = str(i)
            y = y.replace("["," ")
            y = y.replace("]"," ")
            fo.writelines(y)
            fo.writelines("\n")
    else:
        P = str(P)
        fo.writelines(P)
        fo.writelines("\n")
    fo.close()

def get_roi_points(roi_point_3, pointnum):
    '''
    @ Function: 对重复点进行合并
    @ Input:  roi_point_3->(list)  评价后的轮廓段信息
              pointnum->(int)      轮廓点个数
    @ Output: roi_point_3          精选后的分割点   
    @ Author: lxc 
    @ Last modified time: 2022/07/30
    '''
    roi_point_3 = roi_point_3.flatten()
    roi_point_3 = roi_point_3[np.argsort(roi_point_3)]
    Select = [True]*len(roi_point_3)
    for i in range(0,len(roi_point_3),1):
        if abs(roi_point_3[i] - roi_point_3[i-1])<5:
            Select[i] = False
            roi_point_3[i-1] = 0.5* (roi_point_3[i] + roi_point_3[i-1])
        if abs(roi_point_3[i] - roi_point_3[i-1])>pointnum-5:
            Select[i] = False
            roi_point_3[i-1] = 0.5* (roi_point_3[i] + roi_point_3[i-1]-pointnum)
    roi_point_3 = roi_point_3[Select]
    return roi_point_3


def get_circle_param(x, y, P1, P2):
    '''
    @ Function: 获得圆(弧)尺寸信息
    @ Input:  x->(list)            轮廓段横坐标
              y->(list)            轮廓段纵坐标
              P1->(list)           分割点坐标
              P2->(list)           分割点坐标
    @ Output: x1                   圆心横坐标
              y1                   圆心纵坐标
              r                    半径
              circle_param         全部参数信息
    @ Author: lxc 
    @ Last modified time: 2022/09/10
    '''
    x0 = P1[0]
    y0 = P1[1]
    xn = P2[0]
    yn = P2[1]
    if y0!=yn:
        D = ((x0**2-xn**2)+(y0**2-yn**2))/(yn-y0)
        E = (x0-xn)/(yn-y0)
        F = ((yn*x0**2-y0*xn**2)+(yn*y0**2-y0*yn**2))/(y0-yn)
        G = (x0*yn-xn*y0)/(y0-yn)
        H = x+E*y+G
        J = np.sum((x**2+y**2+D*y+F)*H)
        a = -J/np.sum(H**2)
        b = D+a*E
        c = F+a*G
    else:
        n = len(x)
        a = ((x0**2-xn**2)+(y0**2-yn**2))/(xn-x0)
        D = x**2+y**2+a*x
        E = (1/n)*np.sum(D)*np.sum(y)
        F = np.sum(D*y)
        b = (E-F)/(np.sum(y**2)-(1/n)*(np.sum(y))**2)
        c = -np.sum(D+b*y)/n
    a,b,c
    x1=-0.5*a
    y1=-0.5*b
    r = 0.5*math.sqrt(a**2+b**2-4*c)
    circle_param = "圆弧俩端点坐标：("+str(x0)+","+str(y0)+") 、("+str(xn)+","+str(yn)+")"+"\n"+\
                   "圆心坐标     : ("+str(x1)+","+str(y1)+")"+"\n"+\
                   "半径         : "+str(r)+"\n"

    return x1,y1,r,circle_param



def get_line_param(P1, P2):
    '''
    @ Function: 获得直线尺寸信息
    @ Input:  P1->(list)           分割点坐标
              P2->(list)           分割点坐标
    @ Output: angle_               直线倾角
              lenght               直线段长度
              line_param           全部参数信息
    @ Author: lxc 
    @ Last modified time: 2022/09/10
    '''
    #  （y-y1）/(y2-y1)-(x-x1)(x2-x1)=0
    x1 = P1[0]
    y1 = P1[1]
    x2 = P2[0]
    y2 = P2[1]
    angle_ = (P2 - P1)
    angle_ = np.arctan2(np.cross(angle_,np.array([1,0])), 
                            np.dot(angle_,np.array([1,0])))*180/np.pi
    lenght = math.sqrt((x1-x2)**2+(y1-y2)**2)
    line_param = "直线段俩端点坐标： ("+str(x1)+","+str(y1)+") 、("+str(x2)+","+str(y2)+")"+"\n"+\
                 "直线段倾角      :  "+str(angle_)+" "+"\n"+\
                 "直线段长度      :  "+str(lenght)+"\n"
    return angle_, lenght, line_param

def out_results(Points, pointnum, roi_point_3):
    '''
    @ Function: 获得零件全部尺寸信息
    @ Input:  Points->(Array)      轮廓点集合
              pointnum->(int)      轮廓点个数
              roi_point_3->(list)  精选分割点
    @ Output: out                  零件尺寸信息
    @ Author: lxc 
    @ Last modified time: 2022/09/10
    '''
    out = []
    for roi in range(0,len(roi_point_3),1):
        point1 = Points[int(roi_point_3[roi-1])]
        point2 = Points[int(roi_point_3[roi])]
        part = [roi_point_3[roi-1],roi_point_3[roi]]
        x,y=get_x_y(Points, pointnum, part,0)
        x= x.astype(np.float64)
        y= y.astype(np.float64)
        Emax_residu = Fitting_Circle(x,y)    
        Emax_line = Fitting_Line(x,y)
        print("Emax_residu",Emax_residu)
        print("Emax_line",Emax_line)
        if Emax_line-Emax_residu>2:
            x0,y0,r,circle_param = get_circle_param(x, y, point1, point2)
            out.append([x0,y0,r])
            print(circle_param)
        else:
            angle_, lenght, line_param = get_line_param(point1, point2)
            out.append([angle_, lenght])
            print(line_param)
    return out

import XLD

def get_contours_XLD(img):
    '''
    @ Function: 获得零件亚像素轮廓
    @ Input:  img->(Array)         零件图像
    @ Output: contourPoints        亚像素轮廓点集合
              len(contourPoints)   亚像素轮廓点个数
              img_RGB              轮廓图像
    @ Author: lxc 
    @ Last modified time: 2022/09/10
    '''
    kernels_Num, kernels = XLD.kernels_init()
    grayImage, gradients, img_RGB = XLD.img_filter(kernels_Num, kernels, img)
    angle_list, amplitude, angle = XLD.get_agg_amp(kernels_Num, grayImage, gradients)
    edge = XLD.get_single_contour(grayImage, amplitude, angle)
    vPts = XLD.get_contour_DXL(kernels_Num, grayImage, angle_list, amplitude, angle, edge)
    vPts = np.array(vPts)
    order_points= XLD.ordering_contours(vPts)
    contourPoints = np.array(order_points)#.astype(np.float16)
    return contourPoints,len(contourPoints),img_RGB

#测试
if __name__ == '__main__':
    src=cv.imread("C:\\Users\\star\\Desktop\\attachment\\123454.png")      
    # Points, pointnum, src  = get_contours_DXL(src) 
    # print(pointnum)
    Points, pointnum, src = img_init(src)
    AnglesGroup, Angles_Group = get_anglesGroup(Points,int(len(Points[0])/600))
    roi_point, Point2 = get_roi(AnglesGroup, Angles_Group, newPoints = True)
    roi_point = get_pointsgroups(pointnum, roi_point)
    roi_point_2 = get_parts_grade(Points,pointnum, roi_point, 5)
    roi_point_3 = get_preciseROI(pointnum, roi_point_2)
    roi_point_3 = get_roi_points(roi_point_3, pointnum)
    out = out_results(Points, pointnum, roi_point_3)
    #制图
    a=2
    b=3
    for p in roi_point_3:
        color = (random.randint(10,255), random.randint(50,255), random.randint(30,255))
        
        j = Points[int(p)]
        src[int(j[1]-a):int(j[1]+b),int(j[0]-a):int(j[0]+b)]=color
    fig1, ax1 = plt.subplots()
    ax1.imshow(src)#,cmap='gray'
    plt.show()


